import dataclasses
from typing import Union, Dict, List


JSON_TYPE = Union[Dict[str, 'JSON_TYPE'], List['JSON_TYPE'], str, int, float]


@dataclasses.dataclass(frozen=True)
class TaskSummary:
    """A serializable representation of a decoding problem."""
    strong_id: str
    decoder: str
    json_metadata: JSON_TYPE

    def __hash__(self) -> int:
        return self.strong_id.__hash__()


from typing import Optional

import numpy as np
import stim

from sinter.executable_task import ExecutableTask
from sinter.anon_task_stats import AnonTaskStats
from sinter.task_summary import JSON_TYPE
from sinter.existing_data import ExistingData


class Task:
    """A decoding problem and a specification of how to take samples from it."""

    def __init__(self,
                 *,
                 # Information related to what the problem being sampled is.
                 circuit: stim.Circuit,
                 decoder: Optional[str] = None,
                 error_model_for_decoder: Optional[stim.DetectorErrorModel] = None,
                 postselection_mask: Optional[np.ndarray] = None,
                 json_metadata: JSON_TYPE = None,

                 # Information related to how to take samples from the problem.
                 # This is not necessary to specify. If not specified, these
                 # details must be given to `sinter.collect`.
                 max_shots: Optional[int] = None,
                 max_errors: Optional[int] = None,
                 start_batch_size: Optional[int] = None,
                 max_batch_size: Optional[int] = None,
                 max_batch_seconds: Optional[float] = None,
                 previous_stats: AnonTaskStats = AnonTaskStats(),
                 existing_data: Optional[ExistingData] = None,
                 ) -> None:
        """Describes a decoding problem and roughly how to sample it.

        Args:
            circuit: The annotated noisy circuit to sample detection event data
                and logical observable data form.
            decoder: The decoder to use to predict the logical observable data
                from the detection event data. This can be set to None if it
                will be specified later.
            error_model_for_decoder: Defaults to None (automatically derive from
                circuit). The error model to configure the decoder with.
            postselection_mask: Defaults to None (unused). A bit packed bitmask
                identifying detectors that must not fire. Shots where the
                indicated detectors fire are discarded.
            json_metadata: Defaults to None. Custom additional data describing
                the problem. Must be JSON serializable. For example, this could
                be a dictionary with "physical_error_rate" and "code_distance"
                keys.
            max_shots: Defaults to None (unused). Stops the sampling process
                after this many samples have been taken from the circuit.
            max_errors: Defaults to None (unused). Stops the sampling process
                after this many errors have been seen in samples taken from the
                circuit. The actual number sampled errors may be larger due to
                batching.
            start_batch_size: Defaults to None (collector's choice). The very
                first shots taken from the circuit will use a batch of this
                size, and no other batches will be taken in parallel. Once this
                initial fact finding batch is done, batches can be taken in
                parallel and the normal batch size limiting processes take over.
            max_batch_size: Defaults to None (unused). Limits batches from
                taking more than this many shots at once. For example, this can
                be used to ensure memory usage stays below some limit.
            max_batch_seconds: Defaults to None (unused). When set, the recorded
                data from previous shots is used to estimate how much time is
                taken per shot. This information is then used to predict the
                biggest batch size that can finish in under the given number of
                seconds. Limits each batch to be no larger than that.
            previous_stats: If previous information has already been collected,
                it can be specified here. This is useful if you want the manager
                to immediately know roughly what the error rate is and how long
                a batch might take, instead of having to wait for new shots to
                reveal this information again.
            existing_data: If previous information has already been collected,
                it can be specified here.
        """
        if max_shots is not None and max_shots < 0:
            raise ValueError(f'max_shots is not None and max_shots={max_shots} < 0')
        if max_errors is not None and max_errors < 0:
            raise ValueError(f'max_errors is not None and max_errors={max_errors} < 0')
        if start_batch_size is not None and start_batch_size <= 0:
            raise ValueError(f'start_batch_size is not None and start_batch_size={start_batch_size} <= 0')
        if max_batch_size is not None and max_batch_size <= 0:
            raise ValueError(
                f'max_batch_size={max_batch_size} is not None and max_batch_size <= 0')
        if max_batch_seconds is not None and max_batch_seconds <= 0:
            raise ValueError(
                f'max_batch_seconds={max_batch_seconds} is not None and max_batch_seconds <= 0')

        self.circuit = circuit
        self.decoder = decoder
        self.error_model_for_decoder = error_model_for_decoder
        self.postselection_mask = postselection_mask
        self.json_metadata = json_metadata

        self.max_shots = max_shots
        self.max_errors = max_errors
        self.start_batch_size = start_batch_size
        self.max_batch_size = max_batch_size
        self.max_batch_seconds = max_batch_seconds

        self.previous_stats = previous_stats
        if existing_data is not None:
            self.previous_stats += existing_data.stats_for(self.to_case_executable())

    def with_merged_options(self,
             *,
             decoder: Optional[str] = None,
             max_shots: Optional[int] = None,
             max_errors: Optional[int] = None,
             start_batch_size: Optional[int] = None,
             max_batch_size: Optional[int] = None,
             max_batch_seconds: Optional[float] = None,
             existing_data: Optional[ExistingData] = None) -> 'Task':
        return Task(
            circuit=self.circuit,
            error_model_for_decoder=self.error_model_for_decoder,
            postselection_mask=self.postselection_mask,
            json_metadata=self.json_metadata,

            decoder=nullable_single_decoder(self.decoder, decoder),

            max_shots=nullable_min(self.max_shots, max_shots),
            max_errors=nullable_min(self.max_errors, max_errors),
            start_batch_size=nullable_min(self.start_batch_size, start_batch_size),
            max_batch_size=nullable_min(self.max_batch_size, max_batch_size),
            max_batch_seconds=nullable_min(self.max_batch_seconds, max_batch_seconds),
            previous_stats=self.previous_stats,
            existing_data=existing_data,
        )

    def to_case_executable(self) -> ExecutableTask:
        if self.decoder is None:
            raise ValueError('decoder is None')
        return ExecutableTask(
            decoder=self.decoder,
            circuit=self.circuit,
            error_model_for_decoder=self.error_model_for_decoder,
            postselection_mask=self.postselection_mask,
            json_metadata=self.json_metadata,
        )


def nullable_min(a: Optional[int], b: Optional[int]) -> Optional[int]:
    if a is None:
        return b
    if b is None:
        return a
    return min(a, b)


def nullable_single_decoder(a: Optional[str], b: Optional[str]) -> str:
    if a is None and b is None:
        raise ValueError('decoder not specified')
    if a is not None and b is not None:
        raise ValueError('decoder specified to both Task and sinter.collect')
    return a if a is not None else b



import contextlib
import functools
import pathlib
import tempfile
import time
from typing import Optional, Dict, Callable, Tuple

import numpy as np
import stim

from sinter.anon_task_stats import AnonTaskStats
from sinter.decoding_internal import decode_using_internal_decoder
from sinter.decoding_pymatching import decode_using_pymatching


DECODER_METHODS: Dict[str, Callable] = {
    'pymatching': decode_using_pymatching,
    'internal': functools.partial(decode_using_internal_decoder,
                                  use_correlated_decoding=False),
    'internal_correlated': functools.partial(decode_using_internal_decoder,
                                             use_correlated_decoding=True),
}


def _post_select(data: np.ndarray, *, post_mask: Optional[np.ndarray] = None) -> Tuple[int, np.ndarray]:
    if post_mask is None:
        return 0, data
    assert post_mask.shape == (data.shape[1],)
    assert post_mask.dtype == np.uint8
    discarded = np.any(data & post_mask, axis=1)
    num_discards = np.count_nonzero(discarded)
    return num_discards, data[~discarded]


def _split_det_obs_data(*, num_dets: int, num_obs: int, data: np.ndarray):
    if num_obs == 0:
        det_data = data
        obs_data = np.zeros(shape=(det_data.shape[0], 0), dtype=np.bool8)
    else:
        num_det_bytes = (num_dets + 7) // 8
        num_obs_bytes = (num_dets % 8 + num_obs + 7) // 8
        obs_data = data[:, -num_obs_bytes:]
        obs_data = np.unpackbits(obs_data, axis=1, count=num_obs_bytes * 8, bitorder='little')
        obs_data = obs_data[:, num_dets % 8:][:, :num_obs]
        obs_data = obs_data != 0
        det_data = data[:, :num_det_bytes]
        rem = num_dets % 8
        if rem:
            det_data[:, -1] &= np.uint8((1 << rem) - 1)
    assert obs_data.shape[0] == det_data.shape[0]
    assert obs_data.shape[1] == num_obs
    assert det_data.shape[1] == (num_dets + 7) // 8
    return det_data, obs_data


def sample_decode(*,
                  circuit: stim.Circuit,
                  decoder_error_model: stim.DetectorErrorModel,
                  post_mask: Optional[np.ndarray] = None,
                  num_shots: int,
                  decoder: str,
                  tmp_dir: Optional[pathlib.Path] = None) -> AnonTaskStats:
    """Samples how many times a decoder correctly predicts the logical frame.

    Args:
        circuit: The noisy circuit to sample from and decode results for.
        decoder_error_model: The error model to give to the decoder.
        post_mask: Postselection mask. Any samples that have a non-zero result
            at a location where the mask has a 1 bit are discarded. If set to
            None, no postselection is performed.
        num_shots: The number of sample shots to take from the circuit.
        decoder: The name of the decoder to use. Allowed values are:
            "pymatching":
                Use pymatching min-weight-perfect-match decoder.
            "internal":
                Use internal decoder with uncorrelated decoding.
            "internal_correlated":
                Use internal decoder with correlated decoding.
        tmp_dir: An existing directory that is currently empty where temporary
            files can be written as part of performing decoding. If set to
            None, one is created using the tempfile package.
    """
    with contextlib.ExitStack() as exit_stack:
        if tmp_dir is None:
            tmp_dir = exit_stack.enter_context(tempfile.TemporaryDirectory())
        start_time = time.monotonic()

        num_dets = circuit.num_detectors
        num_obs = circuit.num_observables
        assert decoder_error_model.num_detectors == num_dets
        assert decoder_error_model.num_observables == num_obs

        # Sample data using Stim.
        sampler = circuit.compile_detector_sampler()
        concat_data = sampler.sample_bit_packed(num_shots, append_observables=True)

        # Postselect, then split into detection event data and observable data.
        num_discards, concat_data = _post_select(concat_data, post_mask=post_mask)
        det_data, obs_data = _split_det_obs_data(num_dets=num_dets,
                                                 num_obs=num_obs,
                                                 data=concat_data)

        # Perform syndrome decoding to predict observables from detection events.
        decode_method = DECODER_METHODS.get(decoder)
        if decode_method is None:
            raise NotImplementedError(f"Unrecognized decoder: {decoder!r}")
        predictions = decode_method(
            error_model=decoder_error_model,
            bit_packed_det_samples=det_data,
            tmp_dir=tmp_dir,
        )
        assert predictions.shape == obs_data.shape
        assert predictions.dtype == obs_data.dtype == np.bool8

        # Count how many predictions matched the actual observable data.
        num_errors = np.count_nonzero(np.any(predictions != obs_data, axis=1))
        return AnonTaskStats(
            shots=num_shots,
            errors=num_errors,
            discards=num_discards,
            seconds=time.monotonic() - start_time,
        )



import math
from typing import Union

import numpy as np
import pytest

from .probability_util import (
    binary_search, log_binomial, log_factorial, least_squares_output_range, least_squares_slope_range,
    binary_intercept, least_squares_through_point, binomial_relative_likelihood_range,
)


@pytest.mark.parametrize(
    "arg,result",
    {
        0: 0,
        1: 0,
        2: math.log(2),
        3: math.log(2) + math.log(3),
        # These values were taken from wolfram alpha:
        10: 15.1044125730755152952257093292510,
        100: 363.73937555556349014407999336965,
        1000: 5912.128178488163348878130886725,
        10000: 82108.9278368143534553850300635,
        100000: 1051299.2218991218651292781082,
    }.items(),
)
def test_log_factorial(arg, result):
    np.testing.assert_allclose(log_factorial(arg), result, rtol=1e-5)


@pytest.mark.parametrize(
    "n,p,hits,result",
    [
        (1, 0.5, 0, np.log(0.5)),
        (1, 0.5, 1, np.log(0.5)),
        (1, 0.1, 0, np.log(0.9)),
        (1, 0.1, 1, np.log(0.1)),
        (2, [0, 1, 0.1, 0.5], 0, [0, -np.inf, np.log(0.9 ** 2), np.log(0.25)]),
        (2, [0, 1, 0.1, 0.5], 1, [-np.inf, -np.inf, np.log(0.1 * 0.9 * 2), np.log(0.5)]),
        (2, [0, 1, 0.1, 0.5], 2, [-np.inf, 0, np.log(0.1 ** 2), np.log(0.25)]),
        # Magic number comes from PDF[BinomialDistribution[10^10, 10^-6], 10000] on wolfram alpha.
        (10 ** 10, 10 ** -6, 10 ** 4, np.log(0.0039893915536591)),
        # Corner cases.
        (1, 0.0, 0, 0),
        (1, 0.0, 1, -np.inf),
        (1, 1.0, 0, -np.inf),
        (1, 1.0, 1, 0),
        # Array broadcast.
        (2, np.array([0.0, 0.5, 1.0]), 0, np.array([0.0, np.log(0.25), -np.inf])),
        (2, np.array([0.0, 0.5, 1.0]), 1, np.array([-np.inf, np.log(0.5), -np.inf])),
        (2, np.array([0.0, 0.5, 1.0]), 2, np.array([-np.inf, np.log(0.25), 0.0])),
    ],
)
def test_log_binomial(
    n: int, p: Union[float, np.ndarray], hits: int, result: Union[float, np.ndarray]
) -> None:
    np.testing.assert_allclose(log_binomial(n=n, p=p, hits=hits), result, rtol=1e-2)


def test_binary_search():
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=100.1) == 10
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=100) == 10
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=99.9) == 10
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=90) == 9
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=92) == 10
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=-100) == 0
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=10**300) == 10**100


def test_least_squares_through_point():
    fit = least_squares_through_point(
        xs=np.array([1, 2, 3]),
        ys=np.array([2, 3, 4]),
        required_x=1,
        required_y=2)
    np.testing.assert_allclose(fit.slope, 1)
    np.testing.assert_allclose(fit.intercept, 1)

    fit = least_squares_through_point(
        xs=np.array([1, 2, 3]),
        ys=np.array([2, 3, 4]),
        required_x=1,
        required_y=1)
    np.testing.assert_allclose(fit.slope, 1.6, rtol=1e-5)
    np.testing.assert_allclose(fit.intercept, -0.6, atol=1e-5)


def test_binary_intercept():
    t = binary_intercept(func=lambda x: x**2, start_x=5, step=1, target_y=82.3, atol=0.01)
    assert t > 0 and abs(t**2 - 82.3) <= 0.01
    t = binary_intercept(func=lambda x: -x**2, start_x=5, step=1, target_y=-82.3, atol=0.01)
    assert t > 0 and abs(t**2 - 82.3) <= 0.01
    t = binary_intercept(func=lambda x: x**2, start_x=0, step=-1, target_y=82.3, atol=0.01)
    assert t < 0 and abs(t**2 - 82.3) <= 0.01
    t = binary_intercept(func=lambda x: -x**2, start_x=0, step=-1, target_y=-82.3, atol=0.2)
    assert t < 0 and abs(t**2 - 82.3) <= 0.2


def test_least_squares_output_range():
    low, mid, high = least_squares_output_range(
        xs=[1, 2, 3],
        ys=[1, 5, 9],
        target_x=100,
        cost_increase=1,
    )
    assert 300 < low < 390 < mid < 410 < high < 500


def test_least_squares_slope_range():
    low, mid, high = least_squares_slope_range(
        xs=[1, 2, 3],
        ys=[1, 5, 9],
        cost_increase=1,
    )
    np.testing.assert_allclose(mid, 4)
    assert 3 < low < 3.5 < mid < 4.5 < high < 5


def test_likely_error_rate_bounds_shrink_towards_half():
    np.testing.assert_allclose(
        binomial_relative_likelihood_range(num_shots=10 ** 5, num_hits=10 ** 5 / 2, likelihood_ratio=1e-3),
        (0.494122, 0.505878),
        rtol=1e-4,
    )
    np.testing.assert_allclose(
        binomial_relative_likelihood_range(num_shots=10 ** 4, num_hits=10 ** 4 / 2, likelihood_ratio=1e-3),
        (0.481422, 0.518578),
        rtol=1e-4,
    )
    np.testing.assert_allclose(
        binomial_relative_likelihood_range(num_shots=10 ** 4, num_hits=10 ** 4 / 2, likelihood_ratio=1e-2),
        (0.48483, 0.51517),
        rtol=1e-4,
    )
    np.testing.assert_allclose(
        binomial_relative_likelihood_range(num_shots=1000, num_hits=500, likelihood_ratio=1e-3),
        (0.44143, 0.55857),
        rtol=1e-4,
    )
    np.testing.assert_allclose(
        binomial_relative_likelihood_range(num_shots=100, num_hits=50, likelihood_ratio=1e-3),
        (0.3204, 0.6796),
        rtol=1e-4,
    )


@pytest.mark.parametrize("n,c,ratio", [
    (100, 50, 1e-1),
    (100, 50, 1e-2),
    (100, 50, 1e-3),
    (1000, 500, 1e-3),
    (10**6, 100, 1e-3),
    (10**6, 100, 1e-2),
])
def test_likely_error_rate_bounds_vs_log_binomial(n: int, c: int, ratio: float):
    a, b = binomial_relative_likelihood_range(num_shots=n, num_hits=n - c, likelihood_ratio=ratio)

    raw = log_binomial(p=(n - c) / n, n=n, hits=n - c)
    low = log_binomial(p=a, n=n, hits=n - c)
    high = log_binomial(p=b, n=n, hits=n - c)
    np.testing.assert_allclose(
        np.exp(low - raw),
        ratio,
        rtol=1e-2,
    )
    np.testing.assert_allclose(
        np.exp(high - raw),
        ratio,
        rtol=1e-2,
    )



import argparse
import sys
from typing import Iterator, Any, Tuple, Optional, List, Callable

import stim

from sinter.task_stats import TaskStats
from sinter.task import Task
from sinter.collection import collect, post_selection_mask_from_last_detector_coords
from sinter.decoding import DECODER_METHODS
from sinter.main_combine import ExistingData, CSV_HEADER


def iter_file_paths_into_goals(circuit_paths: Iterator[str],
                               metadata_func: Callable,
                               postselect_last_coord_mins: List[Optional[int]],
                               ) -> Iterator[Task]:
    for path in circuit_paths:
        with open(path) as f:
            circuit_text = f.read()
        circuit = stim.Circuit(circuit_text)

        for postselect_last_coord_min in postselect_last_coord_mins:
            post_mask = post_selection_mask_from_last_detector_coords(
                circuit=circuit, last_coord_minimum=postselect_last_coord_min)

            yield Task(
                circuit=circuit,
                postselection_mask=post_mask,
                json_metadata=metadata_func(path=path, circuit=circuit),
            )


def parse_args(args: List[str]) -> Any:
    parser = argparse.ArgumentParser(description='Collect Monte Carlo samples.',
                                     prog='sinter collect')
    parser.add_argument('-circuits',
                        nargs='+',
                        required=True,
                        help='Circuit files to sample from and decode.\n'
                             'This parameter can be given multiple arguments.')
    parser.add_argument('-decoders',
                        choices=sorted(DECODER_METHODS.keys()),
                        nargs='+',
                        required=True,
                        help='How to combine new results with old results.')
    parser.add_argument('-max_shots',
                        type=int,
                        default=None,
                        help='Sampling of a circuit will stop if this many shots have been taken.')
    parser.add_argument('-max_errors',
                        type=int,
                        default=None,
                        help='Sampling of a circuit will stop if this many errors have been seen.')
    parser.add_argument('-processes',
                        required=True,
                        type=int,
                        help='Number of processes to use for simultaneous sampling and decoding.')
    parser.add_argument('-save_resume_filepath',
                        type=str,
                        default=None,
                        help='Activates MERGE mode.\n'
                             "If save_resume_filepath doesn't exist, initializes it with a CSV header.\n"
                             'CSV data already at save_resume_filepath counts towards max_shots and max_errors.\n'
                             'Collected data is appended to save_resume_filepath.\n'
                             'Note that MERGE mode is tolerant to failures: if the process is killed, it can simply be restarted and it will pick up where it left off.\n'
                             'Note that MERGE mode is idempotent: if sufficient data has been collected, no additional work is done when run again.')

    parser.add_argument('-start_batch_size',
                        type=int,
                        default=100,
                        help='Initial number of samples to batch together into one job.\n'
                             'Starting small prevents over-sampling of circuits above threshold.\n'
                             'The allowed batch size increases exponentially from this starting point.')
    parser.add_argument('-max_batch_size',
                        type=int,
                        default=None,
                        help='Maximum number of samples to batch together into one job.\n'
                             'Bigger values increase the delay between jobs finishing.\n'
                             'Smaller values decrease the amount of aggregation of results, increasing the amount of output information.')
    parser.add_argument('-max_batch_seconds',
                        type=int,
                        default=None,
                        help='Limits number of shots in a batch so that the estimated runtime of the batch is below this amount.')
    parser.add_argument('-postselect_last_coord_min',
                        type=int,
                        nargs='+',
                        help='Activates postselection of results. '
                             'Whenever a detecor with a last coordinate at least as large as the given '
                             'value creates a detection event, the shot is discarded.',
                        default=(None,))
    parser.add_argument('-quiet',
                        help='Disables writing progress to stderr.',
                        action='store_true')
    parser.add_argument('-existing_data_filepaths',
                        nargs='*',
                        type=str,
                        default=(),
                        help='CSV data from these files counts towards max_shots and max_errors.\n'
                             'This parameter can be given multiple arguments.')
    parser.add_argument('-metadata_func',
                        type=str,
                        default="{'path': path}",
                        help='A python expression that determines whether a case is kept or not.\n'
                             'Available values:\n'
                             '    path: Relative path to the circuit file, from the command line arguments.\n'
                             '    circuit: The circuit itself, parsed from the file, as a stim.Circuit.\n'
                             'Expected type:\n'
                             '    A value that can be serialized into JSON, like a Dict[str, int].\n'
                             '\n'
                             '    Note that the decoder field is already recorded separately, so storing\n'
                             '    it in the metadata as well would be redundant. But something like\n'
                             '    decoder version could be usefully added.\n'
                             'Examples:\n'
                             '''    -metadata_func "{'path': path}"\n'''
                             '''    -metadata_func "{'n': circuit.num_qubits, 'p': float(path.split('/')[-1].split('.')[0])}"\n'''
                        )
    a = parser.parse_args(args=args)
    for e in a.postselect_last_coord_min:
        if e is not None and e < -1:
            raise ValueError(f'a.postselect_last_coord_min={a.postselect_last_coord_min!r} < -1')
    a.postselect_last_coord_min = [None if e == -1 else e for e in
                                                       a.postselect_last_coord_min]

    a.metadata_func = eval(compile(
        'lambda *, path, circuit: ' + a.metadata_func,
        filename='metadata_func:command_line_arg',
        mode='eval'))
    return a


def open_merge_file(path: str) -> Tuple[Any, ExistingData]:
    try:
        existing = ExistingData.from_file(path)
        return open(path, 'a'), existing
    except FileNotFoundError:
        f = open(path, 'w')
        print(CSV_HEADER, file=f)
        return f, ExistingData()


def main_collect(*, command_line_args: List[str]):
    args = parse_args(args=command_line_args)

    iter_tasks = iter_file_paths_into_goals(
        circuit_paths=args.circuits,
        metadata_func=args.metadata_func,
        postselect_last_coord_mins=args.postselect_last_coord_min,
    )
    num_tasks = len(args.circuits) * len(args.decoders)

    print_to_stdout = args.save_resume_filepath is not None or not args.quiet

    did_work = False

    def on_progress(sample: TaskStats) -> None:
        nonlocal did_work
        if print_to_stdout:
            if not did_work:
                print(CSV_HEADER, flush=True)
            print(sample.to_csv_line(), flush=True)
        did_work = True

    collect(
        num_workers=args.processes,
        hint_num_tasks=num_tasks,
        tasks=iter_tasks,
        print_progress=not args.quiet,
        save_resume_filepath=args.save_resume_filepath,
        existing_data_filepaths=args.existing_data_filepaths,
        progress_callback=on_progress,
        max_errors=args.max_errors,
        max_shots=args.max_shots,
        decoders=args.decoders,
        max_batch_seconds=args.max_batch_seconds,
        max_batch_size=args.max_batch_size,
        start_batch_size=args.start_batch_size,
    )

    if not did_work and not args.quiet:
        print("No work to do.", file=sys.stderr)



import dataclasses
import multiprocessing
import pathlib
import tempfile
from typing import Any
from typing import Optional

from sinter.task_stats import TaskStats
from sinter.executable_task import ExecutableTask
from sinter.anon_task_stats import AnonTaskStats
from sinter.task_summary import TaskSummary


@dataclasses.dataclass
class WorkIn:
    key: Any
    task: ExecutableTask
    summary: TaskSummary
    num_shots: int


@dataclasses.dataclass
class WorkOut:
    key: Any
    sample: Optional[TaskStats]
    error: Optional[BaseException]


def worker_loop(tmp_dir: pathlib.Path,
                inp: multiprocessing.Queue,
                out: multiprocessing.Queue) -> None:
    with tempfile.TemporaryDirectory(dir=tmp_dir) as child_dir:
        while True:
            work: Optional[WorkIn] = inp.get()
            if work is None:
                return
            assert isinstance(work, WorkIn)
            try:
                stats: AnonTaskStats = work.task.sample_stats(
                    num_shots=work.num_shots, tmp_dir=child_dir)
            except BaseException as ex:
                out.put(WorkOut(key=work.key, sample=None, error=ex))
                continue

            sample = TaskStats(
                decoder=work.task.decoder,
                json_metadata=work.summary.json_metadata,
                strong_id=work.summary.strong_id,
                shots=stats.shots,
                discards=stats.discards,
                errors=stats.errors,
                seconds=stats.seconds,
            )
            out.put(WorkOut(sample=sample, key=work.key, error=None))



from sinter.anon_task_stats import (
    AnonTaskStats,
)
from sinter.collection import (
    collect,
    iter_collect,
)
from sinter.csv_out import (
    CSV_HEADER,
)
from sinter.probability_util import (
    binomial_relative_likelihood_range,
)
from sinter.plotting import (
    better_sorted_str_terms,
    plot_discard_rate,
    plot_error_rate,
    DataPointId,
)
from sinter.task import (
    Task,
)
from sinter.task_stats import (
    TaskStats,
)
from sinter.task_summary import (
    TaskSummary,
)



import json
from typing import Any, Dict, Union

import pandas as pd

from sinter.task_stats import TaskStats
from sinter.executable_task import ExecutableTask
from sinter.task_summary import TaskSummary
from sinter.decoding import AnonTaskStats


class ExistingData:
    def __init__(self):
        self.data: Dict[str, TaskStats] = {}

    def stats_for(self, case: Union[ExecutableTask, TaskSummary]) -> AnonTaskStats:
        if isinstance(case, ExecutableTask):
            key = case.strong_id()
        elif isinstance(case, TaskSummary):
            key = case.strong_id
        else:
            raise NotImplementedError(f'{type(case)}')
        if key not in self.data:
            return AnonTaskStats()
        return self.data[key].to_anon_stats()

    def add_sample(self, sample: TaskStats) -> None:
        k = sample.strong_id
        if k in self.data:
            self.data[k] += sample
        else:
            self.data[k] = sample

    def __iadd__(self, other: 'ExistingData') -> 'ExistingData':
        for sample in other.data.values():
            self.add_sample(sample)
        return self

    @staticmethod
    def from_file(path_or_file: Any) -> 'ExistingData':
        frame = pd.read_csv(path_or_file, skipinitialspace=True)
        expected_columns = sorted(["shots",
                                   "discards",
                                   "errors",
                                   "seconds",
                                   "strong_id",
                                   "decoder",
                                   "json_metadata"])
        actual_columns = sorted(frame.columns)
        if actual_columns != expected_columns:
            raise ValueError(
                    f"Bad CSV Data. Expected columns {expected_columns!r} "
                    f"but got {actual_columns!r}.")
        id_to_stats = (frame.
                       groupby(['strong_id', 'decoder', 'json_metadata'], sort=False).
                       sum().
                       to_dict(orient='index'))
        result = ExistingData()
        for (strong_id, decoder, custom_json), row in id_to_stats.items():
            assert strong_id not in result.data
            result.data[strong_id] = TaskStats(
                strong_id=strong_id,
                decoder=decoder,
                json_metadata=json.loads(custom_json),
                shots=row['shots'],
                discards=row['discards'],
                errors=row['errors'],
                seconds=row['seconds'],
            )
        return result



import argparse
from typing import Callable, List, Any, Tuple, Iterable, Optional, TYPE_CHECKING, Union

import matplotlib.pyplot as plt

from sinter.plotting import DataPointId
from sinter.plotting import plot_discard_rate, plot_error_rate
from sinter.task_summary import TaskSummary
from sinter.main_combine import ExistingData

if TYPE_CHECKING:
    import sinter


def parse_args(args: List[str]) -> Any:
    parser = argparse.ArgumentParser(description='Plot collected CSV data.',
                                     prog='sinter plot')
    parser.add_argument('-filter_func',
                        type=str,
                        default="True",
                        help='A python expression that determines whether a case is kept or not.\n'
                             'Available values:\n'
                             '    metadata: The parsed value from the json_metadata column.\n'
                             '    decoder: The decoder that decoded the case.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled.\n'
                             'Expected type:\n'
                             '    Something that can be given to `bool` to get True or False.\n'
                             'Examples:\n'
                             '''    -filter_func "decoder=='pymatching'"\n'''
                             '''    -filter_func "0.001 < metadata['p'] < 0.005"\n''')
    parser.add_argument('-x_func',
                        type=str,
                        default="1",
                        help='A python expression that determines where points go on the x axis.\n'
                             'Available values:\n'
                             '    metadata: The parsed value from the json_metadata column.\n'
                             '    decoder: The decoder that decoded the case.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled.\n'
                             'Expected type:\n'
                             '    Something that can be given to `float` to get a float.\n'
                             'Examples:\n'
                             '''    -x_func "metadata['p']"\n'''
                             '''    -x_func "metadata['path'].split('/')[-1].split('.')[0]"\n'''
                        )
    parser.add_argument('-fig_size',
                        type=int,
                        nargs=2,
                        default=None,
                        help='Desired figure width in pixels.')
    parser.add_argument('-fig_height',
                        type=int,
                        default=None,
                        help='Desired figure height in pixels.')
    parser.add_argument('-group_func',
                        type=str,
                        default="'all data (use -group_func and -x_func to group into curves)'",
                        help='A python expression that determines how points are grouped into curves.\n'
                             'Available values:\n'
                             '    metadata: The parsed value from the json_metadata column.\n'
                             '    decoder: The decoder that decoded the case.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled.\n'
                             'Expected type:\n'
                             '    Something that can be given to `str` to get a useful string.\n'
                             'Examples:\n'
                             '''    -group_func "(decoder, metadata['d'])"\n'''
                             '''    -group_func "metadata['path'].split('/')[-2]"\n'''
                        )
    parser.add_argument('-in',
                        type=str,
                        nargs='+',
                        required=True,
                        help='Input files to get data from.')
    parser.add_argument('-type',
                        choices=['error_rate', 'discard_rate'],
                        nargs='+',
                        default=(),
                        help='Picks the figures to include.')
    parser.add_argument('-out',
                        type=str,
                        default=None,
                        help='Output file to write the plot to.\n'
                             'The file extension determines the type of image.\n'
                             'Either this or -show must be specified.')
    parser.add_argument('-xaxis',
                        type=str,
                        default='[log]',
                        help='Customize the X axis title. Prefix [log] for log scale.')
    parser.add_argument('-show',
                        action='store_true',
                        help='Displays the plot in a window.\n'
                             'Either this or -out must be specified.')
    parser.add_argument('-highlight_likelihood_ratio',
                        type=float,
                        default=1e-3,
                        help='The relative likelihood ratio that determines the color highlights around curves.\n'
                             'Setting this outside (0, 1) disables showing highlights.')

    a = parser.parse_args(args=args)
    if not a.show and a.out is None:
        raise ValueError("Must specify '-out file' or '-show'.")
    a.x_func = eval(compile(
        'lambda *, decoder, metadata, strong_id: ' + a.x_func,
        filename='x_func:command_line_arg',
        mode='eval'))
    a.group_func = eval(compile(
        'lambda *, decoder, metadata, strong_id: ' + a.group_func,
        filename='group_func:command_line_arg',
        mode='eval'))
    a.filter_func = eval(compile(
        'lambda *, decoder, metadata, strong_id: ' + a.filter_func,
        filename='filter_func:command_line_arg',
        mode='eval'))
    return a


def plot(
    *,
    samples: Union[Iterable['sinter.TaskStats'], ExistingData],
    group_func: Callable[[TaskSummary], Any],
    filter_func: Callable[[TaskSummary], Any] = lambda e: True,
    x_func: Callable[[TaskSummary], Any],
    include_discard_rate_plot: bool = False,
    include_error_rate_plot: bool = False,
    highlight_likelihood_ratio: Optional[float] = 1e-3,
    xaxis: str,
    fig_size: Optional[Tuple[int, int]] = None,
) -> Tuple[plt.Figure, List[plt.Axes]]:
    if isinstance(samples, ExistingData):
        total = samples
    else:
        total = ExistingData()
        for sample in samples:
            total.add_sample(sample)
    total.data = {k: v
                  for k, v in total.data.items()
                  if bool(filter_func(v.to_task_summary()))}

    if not include_error_rate_plot and not include_discard_rate_plot:
        include_error_rate_plot = True
        include_discard_rate_plot = any(s.discards for s in total.data.values())

    fig: plt.Figure
    ax_err: Optional[plt.Axes]
    ax_dis: Optional[plt.Axes]
    if include_error_rate_plot and include_discard_rate_plot:
        fig, (ax_err, ax_dis) = plt.subplots(1, 2)
    elif include_error_rate_plot:
        fig, ax_err = plt.subplots(1, 1)
        ax_dis = None
    elif include_discard_rate_plot:
        fig, ax_dis = plt.subplots(1, 1)
        ax_err = None
    else:
        raise NotImplementedError()

    def curve_func(summary: TaskSummary) -> Optional[DataPointId]:
        if not filter_func(summary):
            return None
        return DataPointId(
            curve_label=str(group_func(summary)),
            x=float(x_func(summary)),
        )

    if ax_err is not None:
        plot_error_rate(
            ax=ax_err,
            samples=total.data.values(),
            curve_func=curve_func,
            highlight_likelihood_ratio=highlight_likelihood_ratio,
            xaxis=xaxis,
        )
    if ax_dis is not None:
        plot_discard_rate(
            ax=ax_dis,
            samples=total.data.values(),
            curve_func=curve_func,
            highlight_likelihood_ratio=highlight_likelihood_ratio,
            xaxis=xaxis,
        )

    if fig_size is None:
        fig.set_size_inches(10 * (include_discard_rate_plot + include_error_rate_plot), 10)
        fig.set_dpi(100)
    else:
        w, h = fig_size
        fig.set_size_inches(w / 100, h / 100)
        fig.set_dpi(100)
    fig.tight_layout()
    axs = [e for e in [ax_err, ax_dis] if e is not None]
    return fig, axs


def main_plot(*, command_line_args: List[str]):
    args = parse_args(command_line_args)
    total = ExistingData()
    for file in getattr(args, 'in'):
        total += ExistingData.from_file(file)

    fig, _ = plot(
        samples=total,
        group_func=lambda summary: args.group_func(
            decoder=summary.decoder,
            metadata=summary.json_metadata,
            strong_id=summary.strong_id),
        x_func=lambda summary: args.x_func(
            decoder=summary.decoder,
            metadata=summary.json_metadata,
            strong_id=summary.strong_id),
        filter_func=lambda summary: args.filter_func(
            decoder=summary.decoder,
            metadata=summary.json_metadata,
            strong_id=summary.strong_id),
        include_discard_rate_plot='discard_rate' in args.type,
        include_error_rate_plot='error_rate' in args.type,
        xaxis=args.xaxis,
        fig_size=args.fig_size,
        highlight_likelihood_ratio=args.highlight_likelihood_ratio,
    )
    if args.out is not None:
        fig.savefig(args.out)
    if args.show:
        plt.show()



import math
import pathlib
from typing import Callable, List

import networkx as nx
import numpy as np
import pymatching
import stim


def decode_using_pymatching(*,
                            error_model: stim.DetectorErrorModel,
                            bit_packed_det_samples: np.ndarray,
                            tmp_dir: pathlib.Path,
                            ) -> np.ndarray:
    """Collect statistics on how often logical errors occur when correcting using detections."""

    matching_graph = detector_error_model_to_pymatching_graph(error_model)

    num_shots = bit_packed_det_samples.shape[0]
    num_obs = error_model.num_observables
    num_dets = error_model.num_detectors
    assert bit_packed_det_samples.shape[1] == (num_dets + 7) // 8

    predictions = np.zeros(shape=(num_shots, num_obs), dtype=np.bool8)
    for k in range(num_shots):
        expanded_det = np.unpackbits(bit_packed_det_samples[k], count=num_dets, bitorder='little')
        expanded_det = np.resize(expanded_det, num_dets + 1)
        expanded_det[-1] = 0
        predictions[k] = matching_graph.decode(expanded_det)
    return predictions


def iter_flatten_model(model: stim.DetectorErrorModel,
                       handle_error: Callable[[float, List[int], List[int]], None],
                       handle_detector_coords: Callable[[int, np.ndarray], None]):
    det_offset = 0
    coords_offset = np.zeros(100, dtype=np.float64)

    def _helper(m: stim.DetectorErrorModel, reps: int):
        nonlocal det_offset
        nonlocal coords_offset
        for _ in range(reps):
            for instruction in m:
                if isinstance(instruction, stim.DemRepeatBlock):
                    _helper(instruction.body_copy(), instruction.repeat_count)
                elif isinstance(instruction, stim.DemInstruction):
                    if instruction.type == "error":
                        dets: List[int] = []
                        frames: List[int] = []
                        t: stim.DemTarget
                        p = instruction.args_copy()[0]
                        for t in instruction.targets_copy():
                            if t.is_relative_detector_id():
                                dets.append(t.val + det_offset)
                            elif t.is_logical_observable_id():
                                frames.append(t.val)
                            elif t.is_separator():
                                # Treat each component of a decomposed error as an independent error.
                                # (Ideally we could configure some sort of correlated analysis; oh well.)
                                handle_error(p, dets, frames)
                                frames = []
                                dets = []
                        # Handle last component.
                        handle_error(p, dets, frames)
                    elif instruction.type == "shift_detectors":
                        det_offset += instruction.targets_copy()[0]
                        a = np.array(instruction.args_copy())
                        coords_offset[:len(a)] += a
                    elif instruction.type == "detector":
                        a = np.array(instruction.args_copy())
                        for t in instruction.targets_copy():
                            handle_detector_coords(t.val + det_offset, a + coords_offset[:len(a)])
                    elif instruction.type == "logical_observable":
                        pass
                    else:
                        raise NotImplementedError()
                else:
                    raise NotImplementedError()
    _helper(model, 1)


def detector_error_model_to_nx_graph(model: stim.DetectorErrorModel) -> nx.Graph:
    """Convert a stim error model into a NetworkX graph."""

    g = nx.Graph()
    boundary_node = model.num_detectors
    g.add_node(boundary_node, is_boundary=True, coords=[-1, -1, -1])

    def handle_error(p: float, dets: List[int], frame_changes: List[int]):
        if p == 0:
            return
        if len(dets) == 0:
            # No symptoms for this error.
            # Code probably has distance 1.
            # Accept it and keep going, though of course decoding will probably perform terribly.
            return
        if len(dets) == 1:
            dets = [dets[0], boundary_node]
        if len(dets) > 2:
            raise NotImplementedError(
                f"Error with more than 2 symptoms can't become an edge or boundary edge: {dets!r}.")
        if g.has_edge(*dets):
            edge_data = g.get_edge_data(*dets)
            old_p = edge_data["error_probability"]
            old_frame_changes = edge_data["qubit_id"]
            # If frame changes differ, the code has distance 2; just keep whichever was first.
            if set(old_frame_changes) == set(frame_changes):
                p = p * (1 - old_p) + old_p * (1 - p)
                g.remove_edge(*dets)
        if p > 0.5:
            p = 1 - p
        if p > 0:
            g.add_edge(*dets, weight=math.log((1 - p) / p), qubit_id=frame_changes, error_probability=p)

    def handle_detector_coords(detector: int, coords: np.ndarray):
        g.add_node(detector, coords=coords)

    iter_flatten_model(model, handle_error=handle_error, handle_detector_coords=handle_detector_coords)

    return g


def detector_error_model_to_pymatching_graph(model: stim.DetectorErrorModel) -> pymatching.Matching:
    """Convert a stim error model into a pymatching graph."""
    g = detector_error_model_to_nx_graph(model)
    num_detectors = model.num_detectors
    num_observables = model.num_observables

    # Add spandrels to the graph to ensure pymatching will accept it.
    # - Make sure there's only one connected component.
    # - Make sure no detector nodes are skipped.
    # - Make sure no observable nodes are skipped.
    for k in range(num_detectors):
        g.add_node(k)
    g.add_node(num_detectors + 1)
    for k in range(num_detectors + 1):
        g.add_edge(k, num_detectors + 1, weight=9999999999)
    g.add_edge(num_detectors, num_detectors + 1, weight=9999999999, qubit_id=list(range(num_observables)))

    return pymatching.Matching(g)



import pathlib
import hashlib
import json
import numpy as np
from typing import Any
from typing import Dict
from typing import Optional

import stim

from sinter.anon_task_stats import AnonTaskStats
from sinter.task_summary import JSON_TYPE, TaskSummary


class ExecutableTask:
    """Complete specification of an executable decoding problem."""

    def __init__(self, *,
                 circuit: stim.Circuit,
                 decoder: str,
                 error_model_for_decoder: Optional[stim.DetectorErrorModel] = None,
                 postselection_mask: Optional[np.ndarray] = None,
                 json_metadata: JSON_TYPE = None) -> None:
        if error_model_for_decoder is None:
            error_model_for_decoder = circuit.detector_error_model(decompose_errors=True)
        else:
            n1 = circuit.num_detectors + circuit.num_observables
            n2 = error_model_for_decoder.num_detectors + error_model_for_decoder.num_observables
            assert n1 == n2
            if postselection_mask is not None:
                assert postselection_mask.dtype == np.uint8
                assert postselection_mask.shape == ((n1 + 7) // 8,)
        self.circuit = circuit
        self.decoder_error_model = error_model_for_decoder
        self.decoder = decoder
        self.postselection_mask = postselection_mask
        self.json_metadata = json_metadata

    def to_summary(self) -> TaskSummary:
        return TaskSummary(
            decoder=self.decoder,
            json_metadata=self.json_metadata,
            strong_id=self.strong_id(),
        )

    def strong_id_value(self) -> Dict[str, Any]:
        """Contains all data that affects the strong id for this case.

        This value is converted into the actual strong id by:
            - Serializing it into text using JSON.
            - Serializing the JSON text into bytes using UTF8.
            - Hashing the UTF8 bytes using SHA256.
        """
        return {
            'circuit': str(self.circuit),
            'decoder': self.decoder,
            'decoder_error_model': str(self.decoder_error_model),
            'postselection_mask':
                None
                if self.postselection_mask is None
                else list(self.postselection_mask),
            'json_metadata': self.json_metadata,
        }

    def strong_id_text(self) -> str:
        return json.dumps(self.strong_id_value())

    def strong_id_bytes(self) -> bytes:
        return self.strong_id_text().encode('utf8')

    def strong_id(self) -> str:
        """A cryptographically unique identifier for this task."""
        return hashlib.sha256(self.strong_id_bytes()).hexdigest()

    def sample_stats(self,
                     *,
                     num_shots: int,
                     tmp_dir: Optional[pathlib.Path]) -> AnonTaskStats:
        from sinter.decoding import sample_decode
        return sample_decode(
            num_shots=num_shots,
            circuit=self.circuit,
            post_mask=self.postselection_mask,
            decoder_error_model=self.decoder_error_model,
            decoder=self.decoder,
            tmp_dir=tmp_dir,
        )



import numpy as np
import pytest
import stim

import sinter
from sinter.collection import post_selection_mask_from_last_detector_coords
from sinter.decoding import sample_decode


def test_decode_using_pymatching():
    circuit = stim.Circuit.generated('repetition_code:memory',
                                     rounds=3,
                                     distance=3,
                                     after_clifford_depolarization=0.05)
    result = sample_decode(
        circuit=circuit,
        decoder_error_model=circuit.detector_error_model(decompose_errors=True),
        num_shots=1000,
        decoder='pymatching',
    )
    assert result.discards == 0
    assert 1 <= result.errors <= 100
    assert result.shots == 1000


def test_pymatching_works_on_surface_code():
    circuit = stim.Circuit.generated(
        "surface_code:rotated_memory_x",
        distance=3,
        rounds=15,
        after_clifford_depolarization=0.001,
    )
    stats = sample_decode(
        num_shots=1000,
        circuit=circuit,
        decoder_error_model=circuit.detector_error_model(decompose_errors=True),
        decoder="pymatching",
    )
    assert 0 <= stats.errors <= 50


def test_decode_using_internal_decoder():
    pytest.importorskip('gqec')

    circuit = stim.Circuit.generated('repetition_code:memory',
                                     rounds=3,
                                     distance=3,
                                     after_clifford_depolarization=0.05)
    result = sample_decode(
        circuit=circuit,
        decoder_error_model=circuit.detector_error_model(decompose_errors=True),
        num_shots=1000,
        decoder='internal',
    )
    assert result.discards == 0
    assert 1 <= result.errors <= 100
    assert result.shots == 1000


def test_decode_using_internal_decoder_correlated():
    pytest.importorskip('gqec')

    circuit = stim.Circuit.generated('repetition_code:memory',
                                     rounds=3,
                                     distance=3,
                                     after_clifford_depolarization=0.05)
    result = sample_decode(
        circuit=circuit,
        decoder_error_model=circuit.detector_error_model(decompose_errors=True),
        num_shots=1000,
        decoder='internal_correlated',
    )
    assert result.discards == 0
    assert 1 <= result.errors <= 100
    assert result.shots == 1000


def test_empty():
    circuit = stim.Circuit()
    result = sample_decode(
        circuit=circuit,
        decoder_error_model=circuit.detector_error_model(decompose_errors=True),
        num_shots=1000,
        decoder='pymatching',
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert result.errors == 0


def test_no_observables():
    c = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0
        DETECTOR rec[-1]
    """)
    result = sample_decode(
        circuit=c,
        decoder_error_model=c.detector_error_model(decompose_errors=True),
        num_shots=1000,
        decoder='pymatching',
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert result.errors == 0


@pytest.mark.parametrize("offset", range(8))
def test_observable_offsets_mod8(offset: int):
    c = stim.Circuit("""
        X_ERROR(0.1) 0
        MR 0
        DETECTOR rec[-1]
    """) * (8 + offset) + stim.Circuit("""
        X_ERROR(0.1) 0
        MR 0
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit=c,
        decoder_error_model=c.detector_error_model(decompose_errors=True),
        num_shots=1000,
        decoder='pymatching',
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert 50 <= result.errors <= 150


def test_no_detectors():
    c = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit=c,
        decoder_error_model=c.detector_error_model(decompose_errors=True),
        num_shots=1000,
        decoder='pymatching',
    )
    assert result.discards == 0
    assert 50 <= result.errors <= 150


def test_no_detectors_with_post_mask():
    c = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit=c,
        decoder_error_model=c.detector_error_model(decompose_errors=True),
        post_mask=np.array([0], dtype=np.uint8),
        num_shots=1000,
        decoder='pymatching',
    )
    assert result.discards == 0
    assert 50 <= result.errors <= 150


def test_post_selection():
    c = stim.Circuit("""
        X_ERROR(0.6) 0
        M 0
        DETECTOR(2) rec[-1]
        OBSERVABLE_INCLUDE(0) rec[-1]

        X_ERROR(0.5) 1
        M 1
        DETECTOR(1) rec[-1]
        OBSERVABLE_INCLUDE(0) rec[-1]
        
        X_ERROR(0.1) 2
        M 2
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit=c,
        decoder_error_model=c.detector_error_model(decompose_errors=True),
        post_mask=post_selection_mask_from_last_detector_coords(circuit=c, last_coord_minimum=2),
        num_shots=2000,
        decoder='pymatching',
    )
    assert 1050 <= result.discards <= 1350
    assert 40 <= result.errors <= 160



import collections
import dataclasses
from typing import Callable, TypeVar, List, Any, Tuple, Iterable, DefaultDict, Optional, TYPE_CHECKING

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt

from sinter.anon_task_stats import AnonTaskStats
from sinter.probability_util import binomial_relative_likelihood_range
from sinter.task_summary import TaskSummary

if TYPE_CHECKING:
    import sinter


MARKERS: str = "ov*sp^<>8PhH+xXDd|" * 100
COLORS: Tuple[str, ...] = tuple(mcolors.TABLEAU_COLORS.keys()) * 3


T = TypeVar('T')


def split_by(vs: Iterable[T], key_func: Callable[[T], Any]) -> List[List[T]]:
    cur_key: Any = None
    out: List[List[T]] = []
    buf: List[T] = []
    for item in vs:
        key = key_func(item)
        if key != cur_key:
            cur_key = key
            if buf:
                out.append(buf)
                buf = []
        buf.append(item)
    if buf:
        out.append(buf)
    return out


def better_sorted_str_terms(val: Any) -> Any:
    if isinstance(val, tuple):
        return tuple(better_sorted_str_terms(e) for e in val)
    if not isinstance(val, str):
        return val
    terms = split_by(val, lambda c: c in '.0123456789')
    result = []
    for term in terms:
        term = ''.join(term)
        if '.' in term:
            try:
                term = float(term)
            except ValueError:
                term = tuple(int(e) for e in term.split('.'))
        else:
            try:
                term = int(term)
            except ValueError:
                pass
        result.append(term)
    return tuple(result)


@dataclasses.dataclass
class DataPointId:
    curve_label: str
    x: float
    curve_appearance_id: Optional[Any] = None
    hidden: bool = False

    def effective_appearance_id(self) -> Any:
        if self.curve_appearance_id is not None:
            return self.curve_appearance_id
        return self.curve_label


@dataclasses.dataclass(frozen=True)
class CurveStats:
    label: str
    marker: str
    color: str
    xs: Tuple[float, ...]
    stats: Tuple[AnonTaskStats, ...]
    hidden: bool

    @staticmethod
    def from_samples(samples: Iterable['sinter.TaskStats'],
                     *,
                     curve_func: Callable[['sinter.TaskSummary'], Optional['sinter.DataPointId']],
                     ) -> List['sinter.CurveStats']:
        pairs: DefaultDict[Any, List[Tuple[float, AnonTaskStats]]] = collections.defaultdict(list)

        seen_appearance_ids = set()
        for sample in samples:
            curve = curve_func(sample.to_task_summary())
            if curve is not None:
                aid = curve.effective_appearance_id()
                seen_appearance_ids.add(aid)
                key = curve.curve_label, aid, curve.hidden
                stats = sample.to_anon_stats()
                pairs[key].append((curve.x, stats))

        appearance_id_to_int = {
            k: i
            for i, k in enumerate(sorted(
                seen_appearance_ids,
                key=better_sorted_str_terms,
            ))
        }

        out = []
        for key in sorted(pairs.keys(), key=better_sorted_str_terms):
            val = pairs[key]
            label, appearance_id, hidden = key
            xs = []
            stats = []
            for x, s in sorted(val, key=lambda e: e[0]):
                xs.append(x)
                stats.append(s)
            i = appearance_id_to_int[appearance_id]
            out.append(CurveStats(
                label=label,
                marker=MARKERS[i % len(MARKERS)],
                color=COLORS[i % len(COLORS)],
                xs=tuple(xs),
                stats=tuple(stats),
                hidden=hidden,
            ))

        return out


def plot_discard_rate(
        *,
        ax: plt.Axes,
        samples: 'Iterable[sinter.TaskStats]',
        curve_func: Callable[['sinter.TaskSummary'], Optional['sinter.DataPointId']],
        highlight_likelihood_ratio: Optional[float] = 1e-3,
        xaxis: str,
) -> None:

    ax.set_ylim(0, 1)
    ax.set_ylabel('Discard Probability (per shot)')
    ax.set_yticks([p*0.1 for p in range(11)])
    ax.set_yticklabels([str(p * 10) + '%' for p in range(11)])
    ax.grid()

    if xaxis.startswith('[log]'):
        ax.semilogx()
        xaxis = xaxis[5:]
    if xaxis:
        ax.set_xlabel(xaxis)
        ax.set_title('Discard Rate vs ' + xaxis)
    else:
        ax.set_title('Discard Rates')

    curves = CurveStats.from_samples(samples, curve_func=curve_func)
    for curve in curves:
        if curve.hidden:
            continue
        xs = []
        ys = []
        ys_low = []
        ys_high = []
        for x, stats in zip(curve.xs, curve.stats):
            if stats.shots and highlight_likelihood_ratio is not None:
                xs.append(x)
                ys.append(stats.discards / stats.shots)
                if 0 < highlight_likelihood_ratio < 1:
                    low, high = binomial_relative_likelihood_range(
                        num_shots=stats.shots,
                        num_hits=stats.discards,
                        likelihood_ratio=highlight_likelihood_ratio)
                    ys_low.append(low)
                    ys_high.append(high)
        ax.plot(xs,
                ys,
                label=curve.label,
                marker=curve.marker,
                color=curve.color,
                zorder=100)
        if 0 < highlight_likelihood_ratio < 1:
            ax.fill_between(xs,
                            ys_low,
                            ys_high,
                            color=curve.color,
                            alpha=0.25)

    ax.legend()


def plot_error_rate(
        *,
        ax: plt.Axes,
        samples: 'Iterable[sinter.TaskStats]',
        curve_func: Callable[['sinter.TaskSummary'], Optional['sinter.DataPointId']],
        highlight_likelihood_ratio: Optional[float] = 1e-3,
        xaxis: str = '',
) -> None:
    ax.set_ylabel('Logical Error Probability (per shot)')
    ax.grid()

    if xaxis.startswith('[log]'):
        ax.loglog()
        xaxis = xaxis[5:]
    else:
        ax.semilogy()
    if xaxis:
        ax.set_xlabel(xaxis)
        ax.set_title('Logical Error Rate vs ' + xaxis)
    else:
        ax.set_title('Logical Error Rates')
    all_ys = []

    curves = CurveStats.from_samples(samples, curve_func=curve_func)
    for curve in curves:
        xs = []
        ys = []
        xs_range = []
        ys_low = []
        ys_high = []
        for x, stats in zip(curve.xs, curve.stats):
            num_kept = stats.shots - stats.discards
            if num_kept:
                if stats.errors:
                    xs.append(x)
                    ys.append(stats.errors / num_kept)
                if 0 < highlight_likelihood_ratio < 1:
                    xs_range.append(x)
                    low, high = binomial_relative_likelihood_range(num_shots=num_kept,
                                                                    num_hits=stats.errors,
                                                                    likelihood_ratio=highlight_likelihood_ratio)
                    ys_low.append(low)
                    ys_high.append(high)
        all_ys += ys
        all_ys += ys_low
        all_ys += ys_high
        if not curve.hidden:
            ax.plot(xs,
                    ys,
                    label=curve.label,
                    marker=curve.marker,
                    color=curve.color,
                    zorder=100)
            if 0 < highlight_likelihood_ratio < 1:
                ax.fill_between(xs_range,
                                ys_low,
                                ys_high,
                                color=curve.color,
                                alpha=0.25)

    min_y = min((y for y in all_ys if y > 0), default=0.002)
    low_d = 4
    while 10**-low_d > min_y and low_d < 10:
        low_d += 1
    ax.set_ylim(10**-low_d, 1e-0)
    ax.legend()



import contextlib
import io
import pathlib
import tempfile

import stim

from sinter.main import main
from sinter.main_combine import ExistingData
from sinter.plotting import better_sorted_str_terms, split_by


def test_split_by():
    assert split_by('abcdefcccghi', lambda e: e == 'c') == [list('ab'), list('c'), list('def'), list('ccc'), list('ghi')]


def test_better_sorted_str_terms():
    assert better_sorted_str_terms('a') == ('a',)
    assert better_sorted_str_terms('abc') == ('abc',)
    assert better_sorted_str_terms('a1b2') == ('a', 1, 'b', 2)
    assert better_sorted_str_terms('a1.5b2') == ('a', 1.5, 'b', 2)
    assert better_sorted_str_terms('a1.5.3b2') == ('a', (1, 5, 3), 'b', 2)
    assert sorted([
        "planar d=10 r=30",
        "planar d=16 r=36",
        "planar d=4 r=12",
        "toric d=10 r=30",
        "toric d=18 r=54",
    ], key=better_sorted_str_terms) == [
        "planar d=4 r=12",
        "planar d=10 r=30",
        "planar d=16 r=36",
        "toric d=10 r=30",
        "toric d=18 r=54",
    ]


def test_main_collect():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        for distance in [3, 5, 7]:
            c = stim.Circuit.generated(
                'repetition_code:memory',
                rounds=3,
                distance=distance,
                after_clifford_depolarization=0.02)
            with open(d / f'{distance}.stim', 'w') as f:
                print(c, file=f)

        # Collects requested stats.
        main(command_line_args=[
            "collect",
            "-circuits",
            str(d / "3.stim"),
            str(d / "5.stim"),
            str(d / "7.stim"),
            "-max_shots",
            "1000",
            "-max_errors",
            "10",
            "-decoders",
            "pymatching",
            "-processes",
            "4",
            "-quiet",
            "-save_resume_filepath",
            str(d / "out.csv"),
        ])
        data = ExistingData.from_file(d / "out.csv").data
        assert len(data) == 3
        for k, v in data.items():
            assert v.discards == 0
            assert v.errors <= 50
            assert v.shots >= 1000

        # No more work when existing stats at merge location are sufficient.
        with open(d / "out.csv") as f:
            contents1 = f.read()
        main(command_line_args=[
            "collect",
            "-circuits",
            str(d / "3.stim"),
            str(d / "5.stim"),
            str(d / "7.stim"),
            "-max_shots",
            "1000",
            "-max_errors",
            "10",
            "-decoders",
            "pymatching",
            "-processes",
            "4",
            "-quiet",
            "-save_resume_filepath",
            str(d / "out.csv"),
        ])
        with open(d / "out.csv") as f:
            contents2 = f.read()
        assert contents1 == contents2

        # No more work when existing work is sufficient.
        main(command_line_args=[
            "collect",
            "-circuits",
            str(d / "3.stim"),
            str(d / "5.stim"),
            str(d / "7.stim"),
            "-max_shots",
            "1000",
            "-max_errors",
            "10",
            "-decoders",
            "pymatching",
            "-processes",
            "4",
            "-quiet",
            "-existing_data_filepaths",
            str(d / "out.csv"),
            "-save_resume_filepath",
            str(d / "out2.csv"),
        ])
        data2 = ExistingData.from_file(d / "out2.csv").data
        assert len(data2) == 0


def test_main_combine():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
            """.strip(), file=f)

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "combine",
                str(d / "input.csv"),
            ])
        assert out.getvalue() == """     shots,    errors,  discards, seconds,decoder,strong_id,json_metadata
       600,       101,       220,    3.00,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
         9,         5,         4,    6.00,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
"""

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "combine",
                str(d / "input.csv"),
                str(d / "input.csv"),
            ])
        assert out.getvalue() == """     shots,    errors,  discards, seconds,decoder,strong_id,json_metadata
      1200,       202,       440,    6.00,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
        18,        10,         8,    12.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
"""


def test_main_plot():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
            """.strip(), file=f)

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "plot",
                "-in",
                str(d / "input.csv"),
                "-out",
                str(d / "output.png"),
                "-x_func",
                "int('a' in metadata['path'])",
                "-group_func",
                "decoder",
            ])
        assert (d / "output.png").exists()



import pathlib

import numpy as np
import stim


def decode_using_internal_decoder(*,
                                  error_model: stim.DetectorErrorModel,
                                  bit_packed_det_samples: np.ndarray,
                                  use_correlated_decoding: bool,
                                  tmp_dir: pathlib.Path,
                                  ) -> np.ndarray:
    import gqec  # Internal python wheel.

    num_shots = bit_packed_det_samples.shape[0]
    num_obs = error_model.num_observables
    assert bit_packed_det_samples.shape[1] == (error_model.num_detectors + 7) // 8
    assert bit_packed_det_samples.dtype == np.uint8

    dem_file = f"{tmp_dir}/model.dem"
    dets_file = f"{tmp_dir}/dets.b8"
    out_file = f"{tmp_dir}/predictions.01"

    with open(dem_file, "w") as f:
        print(error_model, file=f)
    with open(dets_file, "wb") as f:
        bit_packed_det_samples.tofile(f)

    gqec.run_finite_match_main(
        dem_filepath=dem_file,
        output_type="predictions",
        output_filepath=out_file,
        dets_has_observables=False,
        dets_format='b8',
        dets_filepath=dets_file,
        ignore_distance_1_errors=True,
        ignore_undecomposed_errors=True,
        use_correlated_decoding=use_correlated_decoding,
    )

    with open(out_file, "rb") as f:
        # Read text data of expected size as if it were binary.
        predictions = np.fromfile(f, dtype=np.uint8)
        predictions.shape = (num_shots, num_obs + 1)

        # Verify newlines are present, then remove them.
        assert np.all(predictions[:, -1] == ord('\n'))
        predictions = predictions[:, :-1]

        # Verify remaining data is 01 data, then convert to bool8.
        assert np.all(ord('0') <= predictions)
        assert np.all(predictions <= ord('1'))
        predictions = predictions == ord('1')

    return predictions



import math
from typing import Union, Callable, Sequence, Tuple

import numpy as np
from scipy.stats import linregress
from scipy.optimize import leastsq
from scipy.stats._stats_mstats_common import LinregressResult


def log_binomial(*, p: Union[float, np.ndarray], n: int, hits: int) -> Union[float, np.ndarray]:
    r"""Approximates $\ln(P(hits = B(n, p)))$; the natural logarithm of a binomial distribution.

    All computations are done in log space to ensure intermediate values can be represented as
    floating point numbers without underflowing to 0 or overflowing to infinity. This is necessary
    when computing likelihoods over many samples. For example, if 80% of a million samples are hits,
    the maximum likelihood estimate is p=0.8. But even this optimal estimate assigns a prior
    probability of roughly 10^-217322 for seeing *exactly* 80% hits out of a million (whereas the
    smallest representable double is roughly 10^-324).

    This method can be broadcast over multiple hypothesis probabilities.

    Args:
        p: The independent probability of a hit occurring for each sample. This can also be an array
            of probabilities, in which case the function is broadcast over the array.
        n: The number of samples that were taken.
        hits: The number of hits that were observed amongst the samples that were taken.

    Returns:
        $\ln(P(hits = B(n, p)))$
    """
    # Clamp probabilities into the valid [0, 1] range (in case float error put them outside it).
    p_clipped = np.clip(p, 0, 1)

    result = np.zeros(shape=p_clipped.shape, dtype=np.float32)
    misses = n - hits

    # Handle p=0 and p=1 cases separately, to avoid arithmetic warnings.
    if hits:
        result[p_clipped == 0] = -np.inf
    if misses:
        result[p_clipped == 1] = -np.inf

    # Multiply p**hits and (1-p)**misses onto the total, in log space.
    result[p_clipped != 0] += np.log(p_clipped[p_clipped != 0]) * hits
    result[p_clipped != 1] += np.log1p(-p_clipped[p_clipped != 1]) * misses

    # Multiply (n choose hits) onto the total, in log space.
    log_n_choose_hits = log_factorial(n) - log_factorial(misses) - log_factorial(hits)
    result += log_n_choose_hits

    return result


def log_factorial(n: int) -> float:
    r"""Approximates $\ln(n!)$; the natural logarithm of a factorial.

    Uses Stirling's approximation for large n.
    """
    if n < 20:
        return sum(math.log(k) for k in range(1, n + 1))
    return (n + 0.5) * math.log(n) - n + math.log(2 * np.pi) / 2


def binary_search(*, func: Callable[[int], float], min_x: int, max_x: int, target: float) -> int:
    """Performs an approximate granular binary search over a monotonically ascending function."""
    while max_x > min_x + 1:
        med_x = (min_x + max_x) // 2
        out = func(med_x)
        if out < target:
            min_x = med_x
        elif out > target:
            max_x = med_x
        else:
            return med_x
    fmax = func(max_x)
    fmin = func(min_x)
    dmax = 0 if fmax == target else fmax - target
    dmin = 0 if fmin == target else fmin - target
    return max_x if abs(dmax) < abs(dmin) else min_x


def binary_intercept(*, func: Callable[[float], float], start_x: float, step: float, target_y: float, atol: float) -> float:
    """Performs an approximate granular binary search over a monotonically ascending function."""
    start_y = func(start_x)
    if abs(start_y - target_y) <= atol:
        return start_x
    while (func(start_x + step) >= target_y) == (start_y >= target_y):
        step *= 2
        if np.isinf(step) or step == 0:
            raise ValueError("Failed.")
    xs = [start_x, start_x + step]
    min_x = min(xs)
    max_x = max(xs)
    increasing = func(min_x) < func(max_x)

    while True:
        med_x = (min_x + max_x) / 2
        med_y = func(med_x)
        if abs(med_y - target_y) <= atol:
            return med_x
        assert med_x not in [min_x, max_x]
        if (med_y < target_y) == increasing:
            min_x = med_x
        else:
            max_x = med_x


def least_squares_cost(*, xs: np.ndarray, ys: np.ndarray, intercept: float, slope: float) -> float:
    assert len(xs.shape) == 1
    assert xs.shape == ys.shape
    return np.sum((intercept + slope*xs - ys)**2)


def least_squares_through_point(*, xs: np.ndarray, ys: np.ndarray, required_x: float, required_y: float) -> LinregressResult:
    xs2 = xs - required_x
    ys2 = ys - required_y

    def err(slope: float) -> float:
        return least_squares_cost(xs=xs2, ys=ys2, intercept=0, slope=slope)

    (best_slope,), _ = leastsq(func=err, x0=0.0)
    intercept = required_y - required_x * best_slope
    return LinregressResult(best_slope, intercept, None, None, None, intercept_stderr=False)


def least_squares_with_slope(*, xs: np.ndarray, ys: np.ndarray, required_slope: float) -> LinregressResult:
    def err(intercept: float) -> float:
        return least_squares_cost(xs=xs, ys=ys, intercept=intercept, slope=required_slope)

    (best_intercept,), _ = leastsq(func=err, x0=0.0)
    return LinregressResult(required_slope, best_intercept, None, None, None, intercept_stderr=False)


def least_squares_output_range(*,
                               xs: Sequence[float],
                               ys: Sequence[float],
                               target_x: float,
                               cost_increase: float) -> Tuple[float, float, float]:
    xs = np.array(xs, dtype=np.float64)
    ys = np.array(ys, dtype=np.float64)
    fit = linregress(xs, ys)
    base_cost = least_squares_cost(xs=xs, ys=ys, intercept=fit.intercept, slope=fit.slope)
    base_y = float(fit.intercept + target_x * fit.slope)

    def cost_for_y(y2: float) -> float:
        fit2 = least_squares_through_point(xs=xs, ys=ys, required_x=target_x, required_y=y2)
        return least_squares_cost(xs=xs, ys=ys, intercept=fit2.intercept, slope=fit2.slope)

    low_y = binary_intercept(start_x=base_y, step=-1, target_y=base_cost + cost_increase, func=cost_for_y, atol=1e-5)
    high_y = binary_intercept(start_x=base_y, step=1, target_y=base_cost + cost_increase, func=cost_for_y, atol=1e-5)
    return low_y, base_y, high_y


def least_squares_slope_range(*,
                              xs: Sequence[float],
                              ys: Sequence[float],
                              cost_increase: float) -> Tuple[float, float, float]:
    xs = np.array(xs, dtype=np.float64)
    ys = np.array(ys, dtype=np.float64)
    fit = linregress(xs, ys)
    base_cost = least_squares_cost(xs=xs, ys=ys, intercept=fit.intercept, slope=fit.slope)

    def cost_for_slope(slope: float) -> float:
        fit2 = least_squares_with_slope(xs=xs, ys=ys, required_slope=slope)
        return least_squares_cost(xs=xs, ys=ys, intercept=fit2.intercept, slope=fit2.slope)

    low_slope = binary_intercept(start_x=fit.slope, step=-1, target_y=base_cost + cost_increase, func=cost_for_slope, atol=1e-5)
    high_slope = binary_intercept(start_x=fit.slope, step=1, target_y=base_cost + cost_increase, func=cost_for_slope, atol=1e-5)
    return low_slope, fit.slope, high_slope


def binomial_relative_likelihood_range(
        *,
        num_shots: int,
        num_hits: int,
        likelihood_ratio: float) -> Tuple[float, float]:
    """Compute the (min, max) probabilities within the given ratio of the max likelihood hypothesis.

    Args:
        num_shots: The number of samples that were taken.
        num_hits: The number of hits that were seen in the samples.
        likelihood_ratio: Should be larger than 1. The maximum Bayes factor between hypotheses and
            the max likelihood hypothesis.

    Returns:
        A (min_hypothesis_probability, max_hypothesis_probability) tuple.
    """
    if num_shots == 0:
        return 0, 1
    log_max_likelihood = log_binomial(p=num_hits / num_shots, n=num_shots, hits=num_hits)
    target_log_likelihood = log_max_likelihood + math.log(likelihood_ratio)
    acc = 100
    low = binary_search(
        func=lambda exp_err: log_binomial(p=exp_err / (acc * num_shots), n=num_shots, hits=num_hits),
        target=target_log_likelihood,
        min_x=0,
        max_x=num_hits * acc) / acc
    high = binary_search(
        func=lambda exp_err: -log_binomial(p=exp_err / (acc * num_shots), n=num_shots, hits=num_hits),
        target=-target_log_likelihood,
        min_x=num_hits * acc,
        max_x=num_shots * acc) / acc
    return low / num_shots, high / num_shots



import contextlib
import multiprocessing
import pathlib
import tempfile
import time
from typing import Iterable
from typing import Optional, Iterator, Tuple, Dict, List

from sinter.existing_data import ExistingData
from sinter.task_stats import TaskStats
from sinter.task import Task
from sinter.collection_tracker_for_single_task import CollectionTrackerForSingleTask
from sinter.worker import worker_loop, WorkIn, WorkOut


class CollectionWorkManager:
    def __init__(self, *,
                 tasks: Iterator[Task],
                 max_shots: Optional[int],
                 max_errors: Optional[int],
                 max_batch_seconds: Optional[int],
                 start_batch_size: Optional[int],
                 max_batch_size: Optional[int],
                 additional_existing_data: Optional[ExistingData],
                 decoders: Optional[Iterable[str]]):
        self.queue_from_workers: Optional[multiprocessing.Queue] = None
        self.queue_to_workers: Optional[multiprocessing.Queue] = None
        self.additional_existing_data = ExistingData() if additional_existing_data is None else additional_existing_data
        self.tmp_dir: Optional[pathlib.Path] = None
        self.exit_stack: Optional[contextlib.ExitStack] = None

        self.max_errors = max_errors
        self.max_shots = max_shots
        self.max_batch_seconds = max_batch_seconds
        self.start_batch_size = start_batch_size
        self.max_batch_size = max_batch_size
        self.decoders = (None,) if decoders is None else tuple(decoders)

        self.workers: List[multiprocessing.Process] = []
        self.active_collectors: Dict[int, CollectionTrackerForSingleTask] = {}
        self.tasks: Iterator[Task] = tasks
        self.next_collector_key: int = 0
        self.finished_count = 0
        self.deployed_jobs: Dict[int, WorkIn] = {}
        self.next_job_id = 0

    def start_workers(self, num_workers: int) -> None:
        assert self.tmp_dir is not None
        current_method = multiprocessing.get_start_method()
        try:
            # To ensure the child processes do not accidentally share ANY state
            # related to, we use 'spawn' instead of 'fork'.
            multiprocessing.set_start_method('spawn', force=True)
            # Create queues after setting start method to work around a deadlock
            # bug that occurs otherwise.
            self.queue_from_workers = multiprocessing.Queue()
            self.queue_to_workers = multiprocessing.Queue()
            self.queue_from_workers.cancel_join_thread()
            self.queue_to_workers.cancel_join_thread()

            for _ in range(num_workers):
                w = multiprocessing.Process(
                    target=worker_loop,
                    args=(self.tmp_dir, self.queue_to_workers, self.queue_from_workers))
                w.start()
                self.workers.append(w)
        finally:
            multiprocessing.set_start_method(current_method, force=True)

    def __enter__(self):
        self.exit_stack = contextlib.ExitStack().__enter__()
        self.tmp_dir = self.exit_stack.enter_context(tempfile.TemporaryDirectory())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shut_down_workers()
        self.exit_stack.__exit__(exc_type, exc_val, exc_tb)
        self.exit_stack = None
        self.tmp_dir = None

    def shut_down_workers(self) -> None:
        removed_workers = self.workers
        self.workers = []

        # Look, we don't have all day here.
        max_shutdown_wait_seconds = 0.1
        deadline = time.monotonic() + max_shutdown_wait_seconds

        # Notify workers that they should stop.
        for _ in removed_workers:
            self.queue_to_workers.put(None)

        # Ensure the workers are stopped.
        for w in removed_workers:
            # Wait until the deadline or until the worker stops.
            max_wait_seconds = deadline - time.monotonic()
            if max_wait_seconds > 0:
                w.join(timeout=max_wait_seconds)
            # Bring the hammer down.
            if w.is_alive():
                w.terminate()

    def fill_work_queue(self) -> bool:
        while len(self.deployed_jobs) < len(self.workers):
            work = self.provide_more_work()
            if work is None:
                break
            self.queue_to_workers.put(WorkIn(key=(self.next_job_id, work.key),
                                             task=work.task,
                                             summary=work.summary,
                                             num_shots=work.num_shots))
            self.deployed_jobs[self.next_job_id] = work
            self.next_job_id += 1
        return bool(self.deployed_jobs)

    def wait_for_next_sample(self,
                             *,
                             timeout: Optional[float] = None,
                             ) -> TaskStats:
        result = self.queue_from_workers.get(timeout=timeout)
        assert isinstance(result, WorkOut)
        if result.error is not None:
            raise RuntimeError("Worker failed") from result.error
        elif result.sample is not None:
            job_id, sub_key = result.key
            self.work_completed(WorkOut(key=sub_key,
                                        sample=result.sample,
                                        error=result.error))
            del self.deployed_jobs[job_id]
            return result.sample
        else:
            raise NotImplementedError(f'result={result!r}')

    def _iter_draw_collectors(self, *, prefer_started: bool) -> Iterator[Tuple[int, CollectionTrackerForSingleTask]]:
        if prefer_started:
            yield from self.active_collectors.items()
        while True:
            key = self.next_collector_key
            try:
                task = next(self.tasks)
            except StopIteration:
                break
            for decoder in self.decoders:
                collector = CollectionTrackerForSingleTask(
                    task=task.with_merged_options(
                        decoder=decoder,
                        max_shots=self.max_shots,
                        max_errors=self.max_errors,
                        max_batch_size=self.max_batch_size,
                        max_batch_seconds=self.max_batch_seconds,
                        start_batch_size=self.start_batch_size,
                        existing_data=self.additional_existing_data,
                    ),
                )
                if collector.is_done():
                    self.finished_count += 1
                    continue
                self.next_collector_key += 1
                self.active_collectors[key] = collector
                yield key, collector
        if not prefer_started:
            yield from self.active_collectors.items()

    def is_done(self) -> bool:
        return len(self.active_collectors) == 0

    def work_completed(self, result: WorkOut):
        collector_index = result.key
        assert isinstance(collector_index, int)
        collector = self.active_collectors[collector_index]
        collector.work_completed(result)
        if collector.is_done():
            self.finished_count += 1
            del self.active_collectors[collector_index]

    def provide_more_work(self) -> Optional[WorkIn]:
        iter_collectors = self._iter_draw_collectors(
                prefer_started=len(self.active_collectors) >= 2)
        for desperate in False, True:
            for collector_index, collector in iter_collectors:
                w = collector.provide_more_work(desperate=desperate)
                if w is not None:
                    assert w.key is None
                    return WorkIn(key=collector_index,
                                  task=w.task,
                                  summary=w.summary,
                                  num_shots=w.num_shots)
        return None

    def status(self, *, num_circuits: Optional[int]) -> str:
        main_status = '\033[31m'
        if num_circuits is not None:
            main_status += f'{num_circuits - self.finished_count} cases not finished yet'
        else:
            main_status += "Running..."
        collector_statuses = [
            '\n    ' + collector.status()
            for collector in self.active_collectors.values()
        ]
        if len(collector_statuses) > 10:
            collector_statuses = collector_statuses[:10] + ['\n...']
        return main_status + ''.join(collector_statuses) + '\033[0m'
